Package: GenProSeq
Type: Package
Title: Generating Protein Sequences with Deep Generative Models
Description: Generative modeling for protein engineering is key to solving fundamental problems in synthetic biology, medicine, and material science. Machine learning has enabled us to generate useful protein sequences on a variety of scales. Generative models are machine learning methods which seek to model the distribution underlying the data, allowing for the generation of novel samples with similar properties to those on which the model was trained. Generative models of proteins can learn biologically meaningful representations helpful for a variety of downstream tasks. Furthermore, they can learn to generate protein sequences that have not been observed before and to assign higher probability to protein sequences that satisfy desired criteria. In this package, common deep generative models for protein sequences, such as variational autoencoder (VAE), generative adversarial networks (GAN), and autoregressive models are available. In the VAE and GAN, the Word2vec is used for embedding. The transformer encoder is applied to protein sequences for the autoregressive model.
Version: 1.16.0
Date: 2024-02-06
Authors@R: c(person(given="Dongmin", family="Jung", email="dmdmjung@gmail.com", role=c("cre", "aut"), comment = c(ORCID = "0000-0001-7499-8422")))
LazyData: FALSE
Depends: keras, mclust, R (>= 4.2)
Imports: tensorflow, word2vec, DeepPINCS, ttgsea, CatEncoders,
        reticulate, stats
Suggests: VAExprs, stringdist, knitr, testthat, rmarkdown
License: Artistic-2.0
biocViews: Software, Proteomics
NeedsCompilation: no
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/GenProSeq
git_branch: RELEASE_3_23
git_last_commit: e5100b8
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
Packaged: 2026-04-28 22:15:58 UTC; biocbuild
Author: Dongmin Jung [cre, aut] (ORCID:
    <https://orcid.org/0000-0001-7499-8422>)
Maintainer: Dongmin Jung <dmdmjung@gmail.com>
Built: R 4.6.0; ; 2026-04-29 04:44:38 UTC; unix
