export("fit_VAE")
export("gen_VAE")
export("fit_GAN")
export("gen_GAN")
export("fit_ART")
export("gen_ART")
export("layer_embedding_token_position")
export("layer_transformer_encoder")
export("prot2vec")
export("vec2prot")
export("prot_seq_check")
import("keras")
import("mclust", except = c("map"))
importFrom("word2vec", "word2vec", "word2vec_similarity")
importFrom("CatEncoders", "LabelEncoder.fit", "transform", "inverse.transform")
importFrom("DeepPINCS", "multiple_sampling_generator", "seq_preprocessing", "get_seq_encode_pad")
importFrom("ttgsea", "sampling_generator")
importFrom("stats", "embed", "predict")
